import doxstractor as dxc


def create_graph(model):
    root_extractor = dxc.CategoryExtractor(
        name="doc_type",
        query="What type of document is this?",
        categories=["lease", "employment"],
        model=model,
    )

    lease_extractor = dxc.TextExtractor(
        name="text_lease", query="What is the content?", model=model
    )
    employment_extractor = dxc.TextExtractor(
        name="text_employment", query="What is the content?", model=model
    )

    children = {
        "lease": [dxc.Node(lease_extractor)],
        "employment": [dxc.Node(employment_extractor)],
    }

    root_node = dxc.Node(root_extractor, children=children)
    return root_node


def test_graph_creation():
    lease_doc = "lease"
    employment_doc = "employment"
    model = dxc.MockModel()

    root_node = create_graph(model)

    expected_lease_result = {"doc_type": "lease", "text_lease": "lease"}
    expected_employment_result = {
        "doc_type": "employment",
        "text_employment": "employment",
    }

    lease_result = root_node.extract(lease_doc)
    employment_result = root_node.extract(employment_doc)

    assert lease_result == expected_lease_result
    assert employment_result == expected_employment_result


def test_graph_creation_with_scores():
    lease_doc = "lease"
    employment_doc = "employment"
    model = dxc.MockModelWithScores()

    root_node = create_graph(model)

    expected_lease_result = {"doc_type": "lease", "text_lease": "lease"}
    expected_employment_result = {
        "doc_type": "employment",
        "text_employment": "employment",
    }

    lease_result = root_node.extract(lease_doc)
    employment_result = root_node.extract(employment_doc)

    assert lease_result == expected_lease_result
    assert employment_result == expected_employment_result



from .base import BaseModel
from typing import Dict, List, Optional
import numpy as np


class MockModel(BaseModel):

    def complete(
        self,
        query: str,
        context: str,
        task_description: Optional[str] = None,
        system_prompt: Optional[str] = None,
    ):
        return context.replace("\n", "")

    def model_description(self):
        return {"type": "text", "scores": False}

    def batch_complete(
        self,
        query: str,
        context: List[str],
        task_description: Optional[str] = None,
        system_prompt: Optional[str] = None,
    ) -> List[str]:
        results = []
        for c in context:
            results.append(self.complete(query=query, context=c))
        return results


class MockModelWithScores(BaseModel):

    def complete(
        self,
        query: str,
        context: str,
        task_description: Optional[str] = None,
        system_prompt: Optional[str] = None,
    ):
        return context.replace("\n", "")

    def model_description(self):
        return {"type": "text", "scores": True}

    def batch_complete_with_scores(
        self,
        query: str,
        context: List[str],
        task_description: Optional[str] = None,
        system_prompt: Optional[str] = None,
    ) -> List[Dict]:

        results = []
        for i, c in enumerate(context):
            res = {
                "score": np.log10(i + 1),
                "answer": self.complete(query=query, context=c),
            }
            results.append(res)
        return results



from .anthropic import AnthropicAPIModel
from .mock import MockModel, MockModelWithScores
from .base import BaseModel
from .transformers_qa import TransformersQAModel
from .transformers_classify import TransformerClassifierModel
from .huggingface_endpoints import HFEndPointQAModel



from ..utils import most_common
from ..models import BaseModel
from .base import BaseExtractor
from typing import List

import numpy as np


TASK_DESCRIPTION = (
    "Valid categories are: {categories_str} \n Use the information below:"
)
SYSTEM_PROMPT = 'Your job is to provide a categorial answer based on provided text. Answer only with the category, and no other text. If there is no relevant information in the text provided, respond with "NA". Do not make things up.'


class CategoryExtractor(BaseExtractor):
    def __init__(
        self,
        name: str,
        query: str,
        categories: List[str],
        model: BaseModel,
        max_chunk_size: float = 10_000,
    ) -> None:
        """Create a new extractor

        Args:
            name (str): A unique name. This identifies the extractor within a graph and provides the attribute name.
            categories (list): Allowed categories.
            model (BaseModel): The natural language model which is used to extract text.
            max_chunk_size (float, optional): Maximum size to chunk data into.. Defaults to 10_000.
        """
        super().__init__(
            name=name,
            query=query,
            max_chunk_size=max_chunk_size,
            model=model,
        )
        self.categories = categories

    def extract(self, doc_text: str) -> str:
        """Extracts a category from a document, similar to zero shot classification.

        Args:
            doc_text (str): The document from which to extract

        Returns:
            str: The extracted category. Guaranteed to be one of the categories provided on init or 'NA'
        """

        merged_chunks = self._chunk_text(doc_text)
        categories_str = "The possible categories are " + ", ".join(
            [f'"{w}"' for w in self.categories]
        )

        # Classifier models don't have actual queries, you just provide the possible categories.
        if self.model.model_description()["type"] == "classifier":
            query = self.categories
        else:
            query = self.query

        if self.model.model_description()["scores"]:
            results_with_scores = self.model.batch_complete_with_scores(
                query=query,
                context=merged_chunks,
                task_description=TASK_DESCRIPTION.format(categories_str=categories_str),
                system_prompt=SYSTEM_PROMPT,
            )
            filtered = [
                r for r in results_with_scores if (r["answer"] in self.categories)
            ]
            all_scores = [r["score"] for r in filtered]
            idx = np.argmax(all_scores)
            consensus = results_with_scores[idx]["answer"]
            return consensus
        else:
            results = self.model.batch_complete(
                query=query,
                context=merged_chunks,
                task_description=TASK_DESCRIPTION.format(categories_str=categories_str),
                system_prompt=SYSTEM_PROMPT,
            )
            valid_answers = [
                m for m in results if (m != "NA") and (m in self.categories)
            ]

            if len(valid_answers) > 0:
                consensus = most_common(valid_answers)
            else:
                consensus = "NA"

            return consensus



from ..models import BaseModel
from typing import List


class BaseExtractor:
    def __init__(
        self,
        name: str,  # Has to be unique within graph as it identifies extractor
        query: str,
        model: BaseModel,
        max_chunk_size: float = 10_000,
    ) -> None:
        """Create a new extractor

        Args:
            name (str): A unique name. This identifies the extractor within a graph and provides the attribute name.
            model (BaseModel): The natural language model which is used to extract text.
            max_chunk_size (float, optional): Maximum size to chunk data into.. Defaults to 10_000.
        """
        # TODO: This maybe should be a model attribute.
        self.max_chunk_size = max_chunk_size
        self.model = model
        self.query = query
        self.name = name

    def _chunk_text(self, doc_text: str) -> List[str]:
        """Splits a document by newlines. Then generates chunks which are shorter than
        max_chunk_size.

        Args:
            doc_text (str): The full document to chunk.

        Returns:
            List[str]: A list of chunks
        """
        chunks = doc_text.split("\n")

        merged_chunks = [""]
        for chunk in chunks:
            prev_chunk = merged_chunks[-1]
            if (len(prev_chunk) + len(chunk)) < (self.max_chunk_size - 1):
                merged_chunks[-1] = prev_chunk + "\n" + chunk
            else:
                merged_chunks.append(chunk)

        return merged_chunks

    def extract(self, doc_text: str):
        raise NotImplementedError



from typing import Optional, List, Dict


class BaseModel:
    def __init__(
        self,
        model: Optional[str] = None,
        temperature: float = 0.0,
        max_tokens: int = 1_000,
    ) -> None:
        self.model = model
        self.temperature = temperature
        self.max_tokens = max_tokens

    def complete(
        self,
        query: str,
        context: str,
        task_description: Optional[str] = None,
        system_prompt: Optional[str] = None,
    ):
        raise NotImplementedError

    def model_description(self):
        raise NotImplementedError

    def batch_complete(
        self,
        query: str,
        context: List[str],
        task_description: Optional[str] = None,
        system_prompt: Optional[str] = None,
    ) -> List[str]:
        raise NotImplementedError

    def batch_complete_with_scores(
        self,
        query: str,
        context: List[str],
        task_description: Optional[str] = None,
        system_prompt: Optional[str] = None,
    ) -> List[Dict]:
        raise NotImplementedError



from .extractors import (
    BaseExtractor,
    NumericExtractor,
    CategoryExtractor,
    TextExtractor,
)
from .nodes import Node

from .models import (
    BaseModel,
    AnthropicAPIModel,
    MockModel,
    MockModelWithScores,
    TransformersQAModel,
    TransformerClassifierModel,
    HFEndPointQAModel,
)



from __future__ import annotations
from .extractors import BaseExtractor, CategoryExtractor
from typing import Dict, Optional, List
import collections


class Node:
    def __init__(
        self, extractor: BaseExtractor, children: Optional[Dict[str, List[Node]]] = None
    ) -> None:
        """A node is an element of a tree which has one or multiple children. Depending on the results of the extractor,
        it recursively calls all the child nodes corresponding to the result.

        E.g. if you have a `CategoryExtractor` with the categories `rental_contract` and `employment_contract`, you likely
        want to extract different attributes depending on what you are dealing with.

        So you would provide a children dictionary as below:
        ```python
        {
            "rental_contract": [
                LeaseDatesExtractor,
                SizeExtractor,
            ],
            "employment_contract": [
                SalaryExtractor,
                StockOptionsExtractor
            ]
        }
        ```

        Args:
            extractor (BaseExtractor): The extractor at the root of the node.
            children (Optional[Dict[str, List[Node]]], optional): Dictionary where key is a category of the
                root node and value is a list of nodes. Defaults to None.
        """

        self.extractor = extractor
        self.children = children
        self.validate()

    def validate(self):
        """Ensures the node is valid."""
        if self.children:
            # Ensure only a category extractor gets children
            if not (isinstance(self.extractor, CategoryExtractor)):
                raise ValueError("Children are only supported for category extractors")

            # Ensure children are only set for categories which exist in the parent
            if not set(self.children.keys()).issubset(set(self.extractor.categories)):
                raise ValueError(
                    "Every child must be a category of the parent extractor"
                )

            # Ensure all children have unique names
            names = node_names(self)
            duplicates = [
                item for item, count in collections.Counter(names).items() if count > 1
            ]
            if len(duplicates) > 0:
                raise ValueError(
                    f"Duplicate node name. All node names need to be unique. Duplicates are {duplicates}"
                )

    def extract(self, doc_text: str) -> Dict:
        """Recursively runs all extractors

        Args:
            doc_text (str): Document text from which to extract.

        Returns:
            Dict: {node_name: node_result}
        """

        # Run Extraction on own extractor
        result = self.extractor.extract(doc_text)
        result_dict = {self.extractor.name: result}

        if self.children and (result in self.children.keys()):
            # Fetch the children relevant to the extracted category
            # We will assume there is a category since we validated that earlier
            # We also assume that categorical extractors only return valid categories.
            child_list = self.children[result]
            for child_node in child_list:
                result_dict.update(child_node.extract(doc_text))
        return result_dict


def node_names(node: Node):
    names = [node.extractor.name]

    if node.children:
        for child_list in node.children.values():
            for child_node in child_list:
                names.extend(node_names(child_node))
    return names



from .base import BaseModel
from transformers import pipeline
from typing import Optional, List, Dict
import torch


class TransformersQAModel(BaseModel):

    def __init__(
        self,
        model: str,
        device: int = 1,
        na_threshold: float = 0.5,
        temperature: float = 0,
        max_tokens: int = 1000,
    ) -> None:
        super().__init__(model, temperature, max_tokens)
        device = 0 if torch.cuda.is_available() else -1
        self.pipeline = pipeline(
            "question-answering", model=model, tokenizer=model, device=device
        )
        self.na_threshold = na_threshold

    def model_description(self):
        return {"type": "text", "scores": True}

    def complete(
        self,
        query: str,
        context: str,
        task_description: Optional[str] = None,
        system_prompt: Optional[str] = None,
    ) -> str:
        """Extracts a snippet of text to answer the query.

        Args:
            query (str): The query specifying what we want to extract.
            context (List[str]): The text from which to extract.
            task_description (Optional[str], optional): Not used, only for compatibility. Defaults to None.
            system_prompt (Optional[str], optional): Not used, only for compatibility. Defaults to None.

        Returns:
            str: The extracted response.
        """
        input = {"question": query, "context": context}
        res = self.pipeline(input)

        if res["score"] >= self.na_threshold:
            return res["answer"]
        else:
            return "NA"

    def batch_complete_with_scores(
        self,
        query: str,
        context: List[str],
        task_description: Optional[str] = None,
        system_prompt: Optional[str] = None,
    ) -> List[Dict]:
        """Extracts a snippet of text to answer the query for each context provided.
        Returns a dictionary including answers and confidence scores.

        Args:
            query (str): The query specifying what we want to extract.
            context (List[str]): The text from which to extract. Model will answer query for each element of list.
            task_description (Optional[str], optional): Not used, only for compatibility. Defaults to None.
            system_prompt (Optional[str], optional): Not used, only for compatibility. Defaults to None.

        Returns:
            List[Dict]: A list of dictionaries like {'score':confidence_score, 'answer':answer_text}
        """
        all_inputs = []
        for c in context:
            all_inputs.append({"question": query, "context": c})

        all_results = self.pipeline(all_inputs)
        return all_results



from ..utils import most_common

from .base import BaseExtractor

import re
import numpy as np

TASK_DESCRIPTION = "Use the information given below."
SYSTEM_PROMPT = 'Your job is to extract a numerical value from a document. Respond with a single number. Do not explain your answer. Do not provide context. If there is no relevant information in the text provided, respond with "NA". Do not make things up.'


class NumericExtractor(BaseExtractor):

    def _numerize(self, answer: str):
        num = "".join(
            re.findall("[-+]?[.]?[\d]+(?:,\d\d\d)*[\.]?\d*(?:[eE][-+]?\d+)?", answer)
        )
        return num

    def extract(self, doc_text: str) -> float:
        """Extracts a number from a document.

        Args:
            doc_text (str): The document text from which to extract.

        Returns:
            float: The extracted number.
        """

        merged_chunks = self._chunk_text(doc_text)

        if self.model.model_description()["scores"]:
            results_with_scores = self.model.batch_complete_with_scores(
                query=self.query,
                context=merged_chunks,
                task_description=TASK_DESCRIPTION,
                system_prompt=SYSTEM_PROMPT,
            )
            filtered = [
                r for r in results_with_scores if (self._numerize(r["answer"]) != "")
            ]
            all_scores = [r["score"] for r in filtered]
            idx = np.argmax(all_scores)
            consensus = self._numerize(results_with_scores[idx]["answer"])
            return consensus
        else:
            results = self.model.batch_complete(
                query=self.query,
                context=merged_chunks,
                task_description=TASK_DESCRIPTION,
                system_prompt=SYSTEM_PROMPT,
            )
            valid_answers = [m for m in results if m != "NA"]

            numerized = []
            for answer in valid_answers:

                num = self._numerize(answer)
                if num != "":
                    numerized.append(num)

            if len(numerized) > 0:
                consensus = most_common(numerized)
            else:
                consensus = "NA"

            return consensus




import re

def parseNumber(text):
    """
        Return the first number in the given text for any locale.
        TODO we actually don't take into account spaces for only
        3-digited numbers (like "1 000") so, for now, "1 0" is 10.
        TODO parse cases like "125,000.1,0.2" (125000.1).

        :example:
        >>> parseNumber("a 125,00 €")
        125
        >>> parseNumber("100.000,000")
        100000
        >>> parseNumber("100 000,000")
        100000
        >>> parseNumber("100,000,000")
        100000000
        >>> parseNumber("100 000 000")
        100000000
        >>> parseNumber("100.001 001")
        100.001
        >>> parseNumber("$.3")
        0.3
        >>> parseNumber(".003")
        0.003
        >>> parseNumber(".003 55")
        0.003
        >>> parseNumber("3 005")
        3005
        >>> parseNumber("1.190,00 €")
        1190
        >>> parseNumber("1190,00 €")
        1190
        >>> parseNumber("1,190.00 €")
        1190
        >>> parseNumber("$1190.00")
        1190
        >>> parseNumber("$1 190.99")
        1190.99
        >>> parseNumber("$-1 190.99")
        -1190.99
        >>> parseNumber("1 000 000.3")
        1000000.3
        >>> parseNumber('-151.744122')
        -151.744122
        >>> parseNumber('-1')
        -1
        >>> parseNumber("1 0002,1.2")
        10002.1
        >>> parseNumber("")

        >>> parseNumber(None)

        >>> parseNumber(1)
        1
        >>> parseNumber(1.1)
        1.1
        >>> parseNumber("rrr1,.2o")
        1
        >>> parseNumber("rrr1rrr")
        1
        >>> parseNumber("rrr ,.o")

    """
    try:
        # First we return None if we don't have something in the text:
        if text is None:
            return None
        if isinstance(text, int) or isinstance(text, float):
            return text
        text = text.strip()
        if text == "":
            return None
        # Next we get the first "[0-9,. ]+":
        n = re.search("-?[0-9]*([,. ]?[0-9]+)+", text).group(0)
        n = n.strip()
        if not re.match(".*[0-9]+.*", text):
            return None
        # Then we cut to keep only 2 symbols:
        while " " in n and "," in n and "." in n:
            index = max(n.rfind(','), n.rfind(' '), n.rfind('.'))
            n = n[0:index]
        n = n.strip()
        # We count the number of symbols:
        symbolsCount = 0
        for current in [" ", ",", "."]:
            if current in n:
                symbolsCount += 1
        # If we don't have any symbol, we do nothing:
        if symbolsCount == 0:
            pass
        # With one symbol:
        elif symbolsCount == 1:
            # If this is a space, we just remove all:
            if " " in n:
                n = n.replace(" ", "")
            # Else we set it as a "." if one occurence, or remove it:
            else:
                theSymbol = "," if "," in n else "."
                if n.count(theSymbol) > 1:
                    n = n.replace(theSymbol, "")
                else:
                    n = n.replace(theSymbol, ".")
        else:
            # Now replace symbols so the right symbol is "." and all left are "":
            rightSymbolIndex = max(n.rfind(','), n.rfind(' '), n.rfind('.'))
            rightSymbol = n[rightSymbolIndex:rightSymbolIndex+1]
            if rightSymbol == " ":
                return parseNumber(n.replace(" ", "_"))
            n = n.replace(rightSymbol, "R")
            leftSymbolIndex = max(n.rfind(','), n.rfind(' '), n.rfind('.'))
            leftSymbol = n[leftSymbolIndex:leftSymbolIndex+1]
            n = n.replace(leftSymbol, "L")
            n = n.replace("L", "")
            n = n.replace("R", ".")
        # And we cast the text to float or int:
        n = float(n)
        if n.is_integer():
            return int(n)
        else:
            return n
    except: pass
    return None

def most_common(lst):
    return max(set(lst), key=lst.count)


import anthropic
from .base import BaseModel
from typing import Optional, List
import time


class AnthropicAPIModel(BaseModel):
    def __init__(
        self,
        model: Optional[str] = "claude-3-haiku-20240307",
        temperature: float = 0.0,
        max_tokens: int = 1_000,
    ) -> None:
        """Model using the Anthropic python API

        Args:
            model (Optional[str], optional): Model name. Defaults to "claude-3-haiku-20240307".
            temperature (float, optional): Defaults to 0.0.
            max_tokens (int, optional):Defaults to 1_000.
        """
        super().__init__(model=model, temperature=temperature, max_tokens=max_tokens)

        self.client = anthropic.Anthropic()

    def model_description(self):
        return {"type": "text", "scores": False}

    def _query_anthropic(self, system_prompt, user_prompt):
        message = self.client.messages.create(
            model=self.model,
            max_tokens=self.max_tokens,
            temperature=self.temperature,
            system=system_prompt,
            messages=[
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "text",
                            "text": user_prompt,
                        }
                    ],
                }
            ],
        )

        return message

    def complete(
        self,
        query: str,
        context: str,
        task_description: Optional[str] = None,
        system_prompt: Optional[str] = None,
    ) -> str:
        """Sends a request to Anthropic and returns the result. Deals with API issues such as overload, timout, etc.

        Args:
            query (str): The query specifying what we want to extract.
            context (str): The text from which to extract.
            task_description (Optional[str], optional): Inserted between the query and the context. Defaults to None.
            system_prompt (Optional[str], optional): System prompt for model. Defaults to None.

        Returns:
            str: Model response text.
        """
        if task_description:
            user_prompt = query + "\n" + task_description + "\n" + context
        else:
            user_prompt = query + "\n" + context

        for attempt in range(10):
            try:
                message = self._query_anthropic(system_prompt, user_prompt)
                return message.content[0].text
            except anthropic.APIConnectionError as e:
                print("The anthropic API could not be reached")
                print(e.__cause__)

            except anthropic.RateLimitError as e:
                print(
                    "A 429 status code was received; sleeping 60s to reset rate limit"
                )
                time.sleep(60)

            except anthropic.APIStatusError as e:
                print("Another non-200-range status code was received")
                print(e.status_code)
                print(e.response)
            else:
                break

        return message.content[0].text

    def batch_complete(
        self,
        query: str,
        context: List[str],
        task_description: Optional[str] = None,
        system_prompt: Optional[str] = None,
    ) -> List[str]:
        """Sends batch of requests to Anthropic and returns the result. Deals with API issues such as overload, timout, etc.

        Args:
            query (str): The query specifying what we want to extract.
            context (List[str]): The text from which to extract. Model will answer query for each element of list.
            task_description (Optional[str], optional): Inserted between the query and the context. Defaults to None.
            system_prompt (Optional[str], optional): System prompt for model. Defaults to None.

        Returns:
            List[str]: Model response text for each context.
        """
        all_results = []
        for c in context:
            result = self.complete(
                query=query,
                context=c,
                task_description=task_description,
                system_prompt=system_prompt,
            )
            all_results.append(result)
        return all_results



from .base import BaseModel
from transformers import pipeline
from typing import Dict, Optional, List
import numpy as np


class TransformerClassifierModel(BaseModel):
    def __init__(
        self, model: str, temperature: float = 0, max_tokens: int = 1000
    ) -> None:
        super().__init__(model, temperature, max_tokens)

        self.classifier = pipeline("zero-shot-classification", model=model)

    def model_description(self):
        return {"type": "classifier", "scores": True}

    def complete(
        self,
        query: List[str],
        context: str,
        task_description: Optional[str] = None,
        system_prompt: Optional[str] = None,
    ):
        res = self.classifier(context, query)
        best_answer_idx = np.argmax(res["scores"])
        return res["labels"][best_answer_idx]

    def batch_complete_with_scores(
        self,
        query: str,
        context: List[str],
        task_description: Optional[str] = None,
        system_prompt: Optional[str] = None,
    ) -> List[Dict]:
        # TODO: Implement this
        return super().batch_complete_with_scores(
            query, context, task_description, system_prompt
        )



from ..utils import most_common
from .base import BaseExtractor
import numpy as np

TASK_DESCRIPTION = "Use the information given below."

SYSTEM_PROMPT = 'Answer only with the relevant text snippet you have found below, and no other text. Do not explain your answer or provide any context. If there is no relevant information in the text provided, respond with "NA". Do not make things up.'


class TextExtractor(BaseExtractor):

    def extract(self, doc_text: str) -> str:
        """Extracts a text snippet.

        Args:
            doc_text (str): The document text from which to extract.

        Returns:
            str: The extracted snippet.
        """
        merged_chunks = self._chunk_text(doc_text)

        if self.model.model_description()["scores"]:
            results_with_scores = self.model.batch_complete_with_scores(
                query=self.query,
                context=merged_chunks,
                task_description=TASK_DESCRIPTION,
                system_prompt=SYSTEM_PROMPT,
            )
            all_scores = [r["score"] for r in results_with_scores]
            idx = np.argmax(all_scores)
            consensus = results_with_scores[idx]["answer"]
            return consensus
        else:
            results = self.model.batch_complete(
                query=self.query,
                context=merged_chunks,
                task_description=TASK_DESCRIPTION,
                system_prompt=SYSTEM_PROMPT,
            )
            valid_answers = []
            valid_indices = []
            for idx, answer in enumerate(results):
                if answer != "NA":
                    valid_answers.append(answer)
                    valid_indices.append(idx)

            if len(valid_answers) > 0:
                consensus = most_common(valid_answers)
            else:
                consensus = "NA"
            return consensus



from .base import BaseModel
from typing import Dict, List, Optional
import requests
import os


class HFEndPointQAModel(BaseModel):
    def __init__(
        self,
        model: str,
        api_token: Optional[str] = None,
        temperature: float = 0,
        max_tokens: int = 1000,
    ) -> None:
        """Huggingface model using hosted endpoints. Requires a valid token.

        Args:
            model (str): The endpoint URL
            api_token (Optional[str], optional): Huggingface token. Defaults to HF_API_TOKEN environment variable.
            temperature (float, optional): For compatibility reasons only. Defaults to 0.
            max_tokens (int, optional): For compatibility reasons only. Defaults to 1000.

        Raises:
            ValueError: If no token is provided.
        """
        super().__init__(model, temperature, max_tokens)

        if api_token:
            self.api_token = api_token
        else:
            env_token = os.environ.get("HF_API_TOKEN")
            if env_token:
                self.api_token = env_token
            else:
                raise ValueError(
                    "The huggingface API requires a token. Either pass an API token or set the HF_API_TOKEN environment variable"
                )

    def model_description(self):
        return {"type": "text", "scores": True}

    def complete(
        self,
        query: str,
        context: str,
        task_description: Optional[str] = None,
        system_prompt: Optional[str] = None,
    ):
        input = {"question": query, "context": context}
        payload = {"inputs": input}

        headers = {"Authorization": f"Bearer {self.api_token}"}
        response = requests.post(self.model, headers=headers, json=payload)
        return response.json()

    def batch_complete_with_scores(
        self,
        query: str,
        context: List[str],
        task_description: Optional[str] = None,
        system_prompt: Optional[str] = None,
    ) -> List[Dict]:
        all_results = []
        for c in context:
            result = self.complete(
                query=query,
                context=c,
                task_description=task_description,
                system_prompt=system_prompt,
            )
            all_results.append(result)
        return all_results
